#!/bin/python

###############################################################################
# An example that creates a NMClient instance for another GMainContext
# and iterates the context while doing an async D-Bus call.
#
# D-Bus is fundamentally async. libnm's NMClient API caches D-Bus objects
# on NetworkManager's D-Bus API. As such, it is "frozen" (with the current
# content of the cache) while not iterating the GMainContext. Only by iterating
# the GMainContext any events are processed and things change.
#
# This means, NMClient heavily uses GMainContext (and GDBusConnection)
# and to operate it, you need to iterate the GMainContext. The synchronous
# API (like NM.Client.new()) is for simple programs but usually not best
# for using NMClient for real applications.
#
# To learn more about GMainContext, read https://developer.gnome.org/documentation/tutorials/main-contexts.html
# When I say "mainloop" or "event loop", I mean GMainContext. GMainLoop is
# a small wrapper around GMainContext to run the context with a boolean
# flag.
#
# Usually, non trivial applications run the GMainContext (or GMainLoop)
# from the main() function and aside some setup and teardown, everything
# happens as events from the event loop.
# This example instead performs synchronous steps, and at the places where
# we need to get the result of some async operation, we iterate the GMainContext
# until we get the result. This may not be how a complex application works,
# but you might do this on a simpler application (like a script) that iterates
# the mainloop whenever it needs to wait for async operations to complete.
#
# Iterating the mainloop might dispatch any other sources that are ready.
# In this example nobody else is scheduling unrelated timers or events, but
# if that happens, your application needs to cope with that.
# E.g. while iterating the mainloop many times, still don't nest running the
# same main context (unless you really know what you do).

###############################################################################

import sys
import time
import traceback

import gi

gi.require_version("NM", "1.0")
from gi.repository import NM, GLib, Gio


###############################################################################


def log(msg=None, prefix=None, suffix="\n"):
    # We use nm_utils_print(), because that uses the same logging
    # mechanism as if you run with "LIBNM_CLIENT_DEBUG=trace". This
    # ensures that messages are in sync.
    if msg is None:
        NM.utils_print(0, "\n")
        return
    if prefix is None:
        prefix = f"[{time.monotonic():.5f}] "
    NM.utils_print(0, f"{prefix}{msg}{suffix}")


def error_is_cancelled(e):
    # Whether error is due to cancellation.
    if isinstance(e, GLib.GError):
        if e.domain == "g-io-error-quark" and e.code == Gio.IOErrorEnum.CANCELLED:
            return True
    return False


###############################################################################


# A Context manager for running a mainloop. Of course, this does
# not do anything magically. You can run the context/mainloop without
# this context object.
#
# This is just to show how we could iterate the GMainContext while waiting
# for an async reply. Note that many non-trivial applications that use glib
# would instead run the mainloop from the main function, only running it once,
# but for the entire duration of the program.
#
# This example and MainLoopRun instead assume that you iterate the maincontext
# for short durations at a time. In particular in this case, where there is
# a dedicated maincontext only for NMClient.
class MainLoopRun:
    def __init__(self, info, ctx, timeout=None):
        self._info = info
        self._loop = GLib.MainLoop(ctx)
        self.cancellable = Gio.Cancellable()
        self._timeout = timeout
        self.got_timeout = False
        self.result = None
        self.error = None
        log(f"MainLoopRun[{self._info}]: create with timeout {self._timeout}")

    def _timeout_cb(self, _):
        log(f"MainLoopRun[{self._info}]: timeout")
        self.got_timeout = True
        self._detach()
        self.cancellable.cancel()
        return False

    def _cancellable_cb(self):
        log(f"MainLoopRun[{self._info}]: cancelled")

    def _detach(self):
        if self._timeout_source is not None:
            self._timeout_source.destroy()
            self._timeout_source = None
        if self._cancellable_id is not None:
            self.cancellable.disconnect(self._cancellable_id)
            self._cancellable_id = None

    def __enter__(self):
        log(f"MainLoopRun[{self._info}]: enter")
        self._timeout_source = None
        if self._timeout is not None:
            self._timeout_source = GLib.timeout_source_new(int(self._timeout * 1000))
            self._timeout_source.set_callback(self._timeout_cb)
            self._timeout_source.attach(self._loop.get_context())
        self._cancellable_id = self.cancellable.connect(self._cancellable_cb)
        self._loop.get_context().push_thread_default()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            # Exception happened.
            log(f"MainLoopRun[{self._info}]: exit with exception")
        else:
            log(f"MainLoopRun[{self._info}]: exit: start mainloop")

            self._loop.run()

            if self.error is not None:
                log(
                    f"MainLoopRun[{self._info}]: exit: complete with error {self.error}"
                )
            elif self.result is not None:
                log(
                    f"MainLoopRun[{self._info}]: exit: complete with result {self.result}"
                )
            else:
                log(f"MainLoopRun[{self._info}]: exit: complete with success")

        self._detach()
        self._loop.get_context().pop_thread_default()
        return False

    def quit(self):
        log(f"MainLoopRun[{self._info}]: quit mainloop")
        self._detach()
        self._loop.quit()


###############################################################################


def get_bus():
    # Let's get the GDBusConnection singleton by calling Gio.bus_get().
    # Since we do everything async, use Gio.bus_get() instead Gio.bus_get_sync().
    with MainLoopRun("get_bus", None, 1) as r:

        def bus_get_cb(source, result, r):
            try:
                c = Gio.bus_get_finish(result)
            except Exception as e:
                r.error = e
            else:
                r.result = c
            r.quit()

        Gio.bus_get(Gio.BusType.SYSTEM, r.cancellable, bus_get_cb, r)

    return r.result


###############################################################################


def create_nmc(dbus_connection):
    # Show how to create and initialize a NMClient asynchronously.
    #
    # NMClient implements GAsyncInitableIface, it thus can be initialized
    # asynchronously. That has actually an advantage, because the sync
    # initialization (GInitableIface) requires to create an internal GMainContext
    # which has an overhead.
    #
    # Also, split the GObject creation and the init_async() call in two.
    # That allows one to pass construct-only parameters, in particular like
    # the instance_flags.

    # Create a separate context for the NMClient. The NMClient is strongly
    # tied to the context used at construct time.
    ctx = GLib.MainContext()
    ctx.push_thread_default()

    log(f"[create_nmc]: use separate context for NMClient: ctx={ctx}")
    try:
        # We create a client asynchronously. There is synchronous
        # NM.Client(), however that requires an internal GMainContext
        # and has thus an overhead. Also, it's obviously blocking.
        #
        # Instead, we initialize it asynchronously, which means
        # we need to iterate the main context. In this case, the
        # context cannot have any other sources dispatched, but
        # if there would be other sources, they might be dispatched
        # while iterating (so this is waiting for the result, but
        # may also dispatch unrelated sources (if any), which you would need
        # to handle).
        #
        # Also, only when using the GObject constructor directly, we can
        # suppress loading the permissions and pass a D-Bus connection.
        nmc = NM.Client(
            instance_flags=NM.ClientInstanceFlags.NO_AUTO_FETCH_PERMISSIONS,
            dbus_connection=dbus_connection,
        )
        log(f"[create_nmc]: new NMClient instance: {nmc}")
    finally:
        # We actually don't need that the ctx is the current thread default
        # later on. NMClient will automatically push it, when necessary.
        ctx.pop_thread_default()

    with MainLoopRun("create_mnc", nmc.get_main_context(), 2) as r:

        def _async_init_cb(nmc, result, r):
            try:
                nmc.init_finish(result)
            except Exception as e:
                log(f"[create_nmc]: init_async() completed with error: {e}")
                r.error = e
            else:
                log(f"[create_nmc]: init_async() completed with success")
            r.quit()

        log(f"[create_nmc]: start init_async()")
        nmc.init_async(GLib.PRIORITY_DEFAULT, r.cancellable, _async_init_cb, r)

    if r.error is None:
        if nmc.get_nm_running():
            log(
                f"[create_nmc]: completed with success (daemon version: {nmc.get_version()}, D-Bus daemon unique name: {nmc.get_dbus_name_owner()})"
            )
        else:
            log(f"[create_nmc]: completed with success (daemon not running)")
        return nmc
    if error_is_cancelled(r.error):
        # Cancelled by us. This happened because we hit the timeout with
        # MainLoopRun.
        log(f"[create_nmc]: failed to initialize within timeout")
        return None
    if not nmc.get_dbus_connection():
        # The NMClient has no D-Bus connection, it usually would try
        # to get one via Gio.bus_get(), but it failed.
        log(f"[create_nmc]: failed to create D-Bus connection: {r.error}")
        return None

    log(f"[create_nmc]: unexpected error creating NMClient ({r.error})")
    # This actually should not happen. There is no other reason why
    # initialization can fail.
    assert False, "NMClient initialization is not supposed to fail"


###############################################################################


def make_call(nmc):
    log("[make_call]: make some async D-Bus call")

    if not nmc:
        log("[make_call]: no NMClient. Skip")
        return

    with MainLoopRun("make_call", nmc.get_main_context(), 1) as r:
        # There are two reasons why async operations are preferable with
        # D-Bus and libnm:
        #
        # - pseudo blocking messes with the ordering of events (see https://smcv.pseudorandom.co.uk/2008/11/nonblocking/).
        # - blocking prevents other things from happening and combining synchronous calls is more limited.
        #
        # So doing async operations is mostly interesting when performing multiple operations in
        # parallel, or when we still want to handle other events while waiting for the reply.
        # The example here does not cover that usage well, because there is only one thing happening.

        def _dbus_call_cb(nmc, result, r):
            try:
                res = nmc.dbus_call_finish(result)
            except Exception as e:
                if error_is_cancelled(e):
                    log(
                        f"[make_call]: dbus_call() completed with cancellation after timeout"
                    )
                else:
                    log(f"[make_call]: dbus_call() completed with error: {e}")

                # I don't understand why, but if you hit this exception (e.g. by setting a low
                # timeout) and pass the exception to the out context, then an additional reference
                # to nmc is leaked, and destroy_nmc() will fail. Workaround
                #
                # r.error = e
                r.error = str(e)
            else:
                log(
                    f"[make_call]: dbus_call() completed with success: {str(res)[:40]}..."
                )
            r.quit()

        log(f"[make_call]: start GetPermissions call")
        nmc.dbus_call(
            NM.DBUS_PATH,
            NM.DBUS_INTERFACE,
            "GetPermissions",
            GLib.Variant.new_tuple(),
            GLib.VariantType("(a{ss})"),
            1000,
            r.cancellable,
            _dbus_call_cb,
            r,
        )

    return r.error is None


###############################################################################


def destroy_nmc(nmc_holder, destroy_mode):
    # The way to shutdown an NMClient is just by unrefing it.
    #
    # At any moment, can an NMClient instance have pending async operations.
    # While unrefing NMClient will cancel them right away, they are only
    # reaped when we iterate the GMainContext some more. That means, if we don't
    # want to leak the GMainContext and the pending operations, we must
    # iterate it some more.
    #
    # To know how much more, there is nmc.get_context_busy_watcher(),
    # We can subscribe a weak reference and keep iterating as long
    # as the watcher is alive.
    #
    # Of course, this only applies if the application wishes to keep running
    # but no longer iterating NMClient's GMainContext. Then you need to ensure
    # that all pending operations in GMainContext are completed (by iterating it).
    #
    # In python, that is a bit tricky, because the caller of destroy_nmc()
    # must give up its reference and pass it here via the @nmc_holder list.
    # You must call destroy_nmc() without having any other reference on
    # nmc.
    #
    # This is just an example. This relies that on this point we only have
    # one reference to NMClient (and it's held by the nmc_holder list).
    # Usually you wouldn't make assumptions about this. Instead, you just
    # assume that you need to keep iterating the GMainContext as long as
    # the context busy watcher is alive, regardless that at this point others
    # might still hold references on the NMClient.

    # Transfer the nmc reference out of the list.
    (nmc,) = nmc_holder
    nmc_holder.clear()

    log(
        f"[destroy_nmc]: destroying NMClient {nmc}: pyref={sys.getrefcount(nmc)}, ref_count={nmc.ref_count}, destroy_mode={destroy_mode}"
    )

    if destroy_mode == 0:
        ctx = nmc.get_main_context()

        finished = []

        def _weak_ref_cb():
            log(f"[destroy_nmc]: context busy watcher is gone")
            finished.clear()
            finished.append(True)

        # We take a weak ref on the context-busy-watcher object and give up
        # our reference on nmc. This must be the last reference, which initiates
        # the shutdown of the NMClient.
        weak_ref = nmc.get_context_busy_watcher().weak_ref(_weak_ref_cb)
        del nmc

        def _timeout_cb(unused):
            if not finished:
                # Somebody else holds a reference to the NMClient and keeps
                # it alive. We cannot properly clean up.
                log(
                    f"[destroy_nmc]: ERROR: timeout waiting for context busy watcher to be gone"
                )
                finished.append(False)
            return False

        timeout_source = GLib.timeout_source_new(1000)
        timeout_source.set_callback(_timeout_cb)
        timeout_source.attach(ctx)

        while not finished:
            log(f"[destroy_nmc]: iterating main context")
            ctx.iteration(True)

        timeout_source.destroy()

        log(f"[destroy_nmc]: done: {finished[0]}")
        if not finished[0]:
            weak_ref.unref()
            raise Exception("Failure to destroy NMClient: something keeps it alive")

    else:
        if destroy_mode == 1:
            ctx = GLib.MainContext.default()
        else:
            # Run the maincontext of the NMClient.
            ctx = nmc.get_main_context()
        with MainLoopRun("destroy_nmc", ctx, 2) as r:

            def _wait_shutdown_cb(source_unused, result, r):
                try:
                    NM.Client.wait_shutdown_finish(result)
                except Exception as e:
                    if error_is_cancelled(e):
                        log(
                            f"[destroy_nmc]: wait_shutdown() completed with cancellation after timeout"
                        )
                    else:
                        log(f"[destroy_nmc]: wait_shutdown() completed with error: {e}")
                else:
                    log(f"[destroy_nmc]: wait_shutdown() completed with success")

                r.quit()

            nmc.wait_shutdown(True, r.cancellable, _wait_shutdown_cb, r)
            del nmc


###############################################################################


def run1():
    try:
        dbus_connection = get_bus()
        log()

        nmc = create_nmc(dbus_connection)
        log()

        make_call(nmc)
        log()

        if not nmc:
            log(f"[destroy_nmc]: nothing to destroy")
        else:
            # To cleanup the NMClient, we need to give up the reference. Move
            # it to a list, and destroy_nmc() will take care of it.
            nmc_holder = [nmc]
            del nmc

            # In the example, there are three modes how the destroy is
            # implemented.
            destroy_nmc(nmc_holder, destroy_mode=1)

        log()
        log("done")
    except Exception as e:
        log()
        log("EXCEPTION:")
        log(f"{e}")
        for tb in traceback.format_exception(None, e, e.__traceback__):
            for l in tb.split("\n"):
                log(f">>> {l}")
        return False
    return True


if __name__ == "__main__":
    if not run1():
        sys.exit(1)



#
# Sphinx Documentation Configuration
#

import re
import os
import sys

import capidocs.kerneldoc
import hawkmoth

# Global Setup

project = 'c-rbtree'

author = 'C-Util Community'
copyright = '2015-2022, C-Util Community'

# Hawkmoth C-Audodoc Setup

capidocs.kerneldoc.hawkmoth_conf()

# Extensions

exclude_patterns = []

extensions = [
    'hawkmoth',
]

# Hawkmoth Options

cautodoc_clang = capidocs.kerneldoc.hawkmoth_include_args()
cautodoc_clang += ["-I" + os.path.abspath("..")]
cautodoc_clang += capidocs.kerneldoc.hawkmoth_glob_includes("../../subprojects", "libc*/src")

cautodoc_root = os.path.abspath('..')

cautodoc_transformations = {
    'kerneldoc': capidocs.kerneldoc.hawkmoth_converter,
}

# HTML Options

html_theme = 'sphinx_rtd_theme'



#!/usr/bin/env python
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2020 Red Hat, Inc.

# This example performs a scan of Wi-Fi P2P peers and connects to one
# of them.

import sys
import uuid
import gi

gi.require_version("NM", "1.0")
from gi.repository import GLib, NM

main_loop = None
client = None


def create_profile(name, peer_mac):
    profile = NM.SimpleConnection.new()

    s_con = NM.SettingConnection.new()
    s_con.set_property(NM.SETTING_CONNECTION_ID, name)
    s_con.set_property(NM.SETTING_CONNECTION_UUID, str(uuid.uuid4()))
    s_con.set_property(NM.SETTING_CONNECTION_TYPE, "wifi-p2p")
    s_con.set_property(NM.SETTING_CONNECTION_AUTOCONNECT, False)

    s_ip4 = NM.SettingIP4Config.new()
    s_ip4.set_property(NM.SETTING_IP_CONFIG_METHOD, "auto")

    s_ip6 = NM.SettingIP6Config.new()
    s_ip6.set_property(NM.SETTING_IP_CONFIG_METHOD, "auto")

    s_wifi_p2p = NM.SettingWifiP2P.new()
    s_wifi_p2p.set_property(NM.SETTING_WIFI_P2P_PEER, peer_mac)
    s_wifi_p2p.set_property(
        NM.SETTING_WIFI_P2P_WFD_IES,
        GLib.Bytes.new(b"\x00\x00\x06\x00\x90\x1c\x44\x00\xc8"),
    )

    profile.add_setting(s_con)
    profile.add_setting(s_ip4)
    profile.add_setting(s_ip6)
    profile.add_setting(s_wifi_p2p)

    return profile


def activated_cb(client, result, data):
    try:
        client.add_and_activate_connection2_finish(result)
        print(" * Connection profile activated successfully")
    except Exception as e:
        sys.stderr.write("Error: %s\n" % e)
    main_loop.quit()


def scan_timeout_cb(device):
    peers = device.get_peers()
    if len(peers) == 0:
        main_loop.quit()
        sys.exit("No peer found")

    print("\n   {:20} {:30} {:3} {:30}".format("MAC", "Name", "Sig", "Wfd-IEs"))
    for p in peers:
        if p.get_wfd_ies() is not None:
            ies = p.get_wfd_ies().get_data().hex()
        else:
            ies = ""
        print(
            "   {:20} {:30} {:3} {:30}".format(
                p.get_hw_address(), p.get_name(), p.get_strength(), ies
            )
        )
    print("")

    # Connect to first peer
    profile = create_profile("P2P-connection", peers[0].get_hw_address())
    client.add_and_activate_connection2(
        profile, device, "/", GLib.Variant("a{sv}", {}), None, activated_cb, None
    )
    print(
        " * Connecting to peer {} using profile '{}'".format(
            peers[0].get_hw_address(), profile.get_id()
        )
    )


def start_find_cb(device, async_result, user_data):
    try:
        device.start_find_finish(async_result)
    except Exception as e:
        sys.stderr.write("Error: %s\n" % e)
        main_loop.quit()

    print(" * Scanning on device {}...".format(device.get_iface()))
    GLib.timeout_add(10000, scan_timeout_cb, device)


if __name__ == "__main__":
    client = NM.Client.new(None)
    device = None

    devices = client.get_devices()
    for d in devices:
        if d.get_device_type() == NM.DeviceType.WIFI_P2P:
            device = d
            break

    if device is None:
        sys.exit("No Wi-Fi P2P device found")

    device.start_find(GLib.Variant("a{sv}", {}), None, start_find_cb, None)

    main_loop = GLib.MainLoop()
    main_loop.run()



#!/usr/bin/env python
# SPDX-License-Identifier: LGPL-2.1-or-later

import gi

gi.require_version("NM", "1.0")
from gi.repository import GLib, NM, Gio

# This example shows how to implement a very simple secret agent for
# NetworkManager. The secret agent registers to the NM daemon and can
# provide missing secrets like Wi-Fi or VPN passwords. Set environment
# variable "LIBNM_CLIENT_DEBUG=trace" to enable libnm verbose logging.


class SecretAgent(NM.SecretAgentOld):
    def __init__(self):
        super().__init__(identifier="MySecretAgent")
        super().init()

    def do_get_secrets(
        self,
        connection,
        connection_path,
        setting_name,
        hints,
        flags,
        callback,
        callback_data,
    ):
        print(
            "get_secrets for '{}', interface '{}', setting '{}'".format(
                connection.get_id(), connection.get_interface_name(), setting_name
            )
        )

        # Implement here the logic to retrieve the secrets.
        # As an example, we return a hardcoded Wi-Fi PSK.
        if (
            connection.get_connection_type() == "802-11-wireless"
            and setting_name == "802-11-wireless-security"
        ):
            s_wifi = connection.get_setting_wireless()
            ssid = NM.utils_ssid_to_utf8(s_wifi.get_ssid().get_data())

            if ssid == "home":
                secrets = GLib.Variant(
                    "a{sa{sv}}",
                    {
                        "802-11-wireless-security": {
                            "psk": GLib.Variant("s", "abcd1234")
                        }
                    },
                )
                print("Sending secrets {}".format(secrets))
                callback(self, connection, secrets, None)
                return

        # We don't have the secret, NM will ask to another agent or fail
        callback(
            self,
            connection,
            None,
            GLib.GError.new_literal(
                NM.SecretAgentError.quark(),
                "No secrets found",
                NM.SecretAgentError.NOSECRETS,
            ),
        )

    def do_cancel_get_secrets(self, connection_path, connection_name):
        pass

    def do_save_secrets(self, connection, connection_path, callback, callback_data):
        # Implement this if you want to store "agent-owned" secrets
        callback(self, connection, None)

    def do_delete_secrets(self, connection, connection_path, callback, callback_data):
        # Implement this if you want to store "agent-owned" secrets
        callback(self, connection, None)


def main():
    agent = SecretAgent()
    loop = GLib.MainLoop()
    try:
        loop.run()
    except KeyboardInterrupt:
        print("Exiting Secret Agent...")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2019 Red Hat, Inc.
#

import gi

gi.require_version("NM", "1.0")
from gi.repository import NM

if __name__ == "__main__":
    client = NM.Client.new(None)
    devices = client.get_devices()

    for d in devices:
        print(
            "{:<16} {:<16} {}".format(
                d.get_iface(),
                "(" + d.get_type_description() + ")",
                NM.utils_enum_to_str(NM.DeviceInterfaceFlags, d.get_interface_flags()),
            )
        )



#!/usr/bin/env python
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import sys

import gi

gi.require_version("NM", "1.0")
from gi.repository import NM, GLib, Gio, GObject


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def kf_from_data(data):
    kf = GLib.KeyFile.new()
    kf.load_from_data(data, 18446744073709551615, GLib.KeyFileFlags.NONE)
    return kf


def kf_to_data(kf):
    data, l = kf.to_data()
    return data


def connection_to_kf(connection):
    return kf_to_data(NM.keyfile_write(connection, NM.KeyfileHandlerFlags.NONE))


def connection_from_kf(data):
    base_dir = os.getcwd()
    return NM.keyfile_read(kf_from_data(data), base_dir, NM.KeyfileHandlerFlags.NONE)


def connection_from_stdin():
    return connection_from_kf(sys.stdin.read())


def device_get_applied_connection(device):
    mainloop = GLib.MainLoop()
    r = []

    def cb(device, result):
        try:
            connection, version_id = device.get_applied_connection_finish(result)
        except Exception as e:
            r.append(e)
        else:
            r.append(connection)
            r.append(version_id)
        mainloop.quit()

    device.get_applied_connection_async(0, None, cb)
    mainloop.run()
    if len(r) == 1:
        raise r[0]
    connection, version_id = r
    return connection, version_id


def device_reapply(device, connection, version_id, reapply_flags):
    mainloop = GLib.MainLoop()
    r = []

    def cb(device, result):
        try:
            device.reapply_finish(result)
        except Exception as e:
            r.append(e)
        mainloop.quit()

    device.reapply_async(connection, version_id or 0, reapply_flags, None, cb)
    mainloop.run()
    if len(r) == 1:
        raise r[0]


def parse_args():
    import argparse

    parser = argparse.ArgumentParser(
        prog="device-reapply.py",
        description="Example program to interact with the applied connection",
    )

    parser.add_argument("mode", choices=["get", "reapply", "modify"])
    parser.add_argument("device")
    parser.add_argument("-V", "--version-id", type=int)
    parser.add_argument("-s", "--stdin", action="store_true")
    parser.add_argument("-p", "--preserve-external-ip", action="store_true")

    return parser.parse_args()


def main():
    args = parse_args()

    nmc = NM.Client.new()

    device = [d for d in nmc.get_devices() if d.get_iface() == args.device]
    if not device:
        raise Exception(f'Device "{args.device}" not found')
    if len(device) != 1:
        raise Exception(f'Not unique device "{args.device}" found')
    device = device[0]

    assert not args.stdin or args.mode == "modify"
    assert not args.preserve_external_ip or args.mode in ["modify", "reapply"]

    if args.mode == "get":
        connection, version_id = device_get_applied_connection(device)

        version_id_matches = args.version_id is None or args.version_id == version_id

        print(
            f'# Applied connection on "{device.get_iface()}": "{connection.get_id()}" ({connection.get_uuid()}, {connection.get_connection_type()})'
        )
        s = "" if version_id_matches else f" (expected {args.version_id})"
        print(f"# version-id={version_id}{s}")
        print(f"#")
        print(f"{connection_to_kf(connection)}")

        if not version_id_matches:
            eprint(
                f"Applied version-id does not match (expects {args.version_id} but got {version_id})"
            )
            sys.exit(1)
        sys.exit(0)

    if args.mode == "reapply":
        new_connection = None
    elif args.stdin:
        new_connection = connection_from_stdin()
    else:
        new_connection, _ = device_get_applied_connection(device)

    reapply_flags = 0
    if args.preserve_external_ip:
        reapply_flags = 1  # NM.DeviceReapplyFlags.PRESERVE_EXTERNAL_IP

    device_reapply(device, new_connection, args.version_id, reapply_flags)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2025 Red Hat, Inc.

from pathlib import Path
import subprocess
import itertools
import shutil
import sys
import os

xgettext = shutil.which("xgettext")

# Paths relative to repo root which should be
# excluded from pot->tree, or tree->pot checks.
ignore_in_potfiles = ()
ignore_in_tree = ()


def err(s):
    print(f"ERR: {s}", file=sys.stderr)


def read_potfile_list(root, path):
    with open(root / path, "r") as f:
        for line in f.readlines():
            stripped = line.strip()
            if stripped.startswith("#") or len(stripped) == 0:
                continue
            if stripped in ignore_in_potfiles:
                continue
            yield (root / stripped).resolve()


def get_existing_entries(root):
    in_list = read_potfile_list(root, "po/POTFILES.in")
    skip_list = read_potfile_list(root, "po/POTFILES.skip")
    return set(itertools.chain(in_list, skip_list))


def check_exists_in_tree(root, paths):
    is_ok = True

    for path in paths:
        if not path.exists():
            err(
                f"{path.relative_to(root)} exists in POTFILES.in or POTFILES.skip, but missing in sources"
            )
            is_ok = False

    return is_ok


def get_gettext_args(root):
    arg_name = "XGETTEXT_OPTIONS ="
    with open(root / "po" / "Makevars", "r") as f:
        for line in f.readlines():
            if line.startswith(arg_name):
                return line[len(arg_name) :].strip().split(" ")

    raise Exception("could not get gettext args")


def list_c_sources(root):
    for path, _, files in os.walk(root / "src"):
        for file in files:
            relpath_str = str(Path(path).relative_to(root) / file)
            extension = file.replace(".in", "").split(".").pop()
            if extension in ["c", "h"] and relpath_str not in ignore_in_tree:
                full_path = root / path / file
                yield full_path.resolve()


def gettext_dry_run(root, paths):
    args = get_gettext_args(root)
    process = subprocess.Popen(
        [xgettext, "--files-from=-", "--output=-"] + args,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        text=True,
    )

    for line in paths:
        process.stdin.write(f"{line.resolve()}\n")
    process.stdin.close()

    seen = set()
    for line in process.stdout:
        if line.startswith("#: "):
            out_path = Path(line.replace("#: ", "").split(":")[0])
            if out_path not in seen:
                seen.add(out_path)
                yield out_path

    process.stdout.close()
    assert process.wait() == 0


def check_exists_in_potfiles(root, pot_paths):
    unseen_paths = filter(lambda path: path not in pot_paths, list_c_sources(root))

    is_ok = True
    for path in gettext_dry_run(root, unseen_paths):
        err(
            f"{path.relative_to(root)} code contains gettext macros, but missing in POTFILES.in or POTFILES.skip"
        )
        is_ok = False

    return is_ok


def check_potfiles():
    root = (Path(os.path.dirname(__file__)) / "../../").resolve()
    file_entries = get_existing_entries(root)
    is_ok = True

    # Let's first check that all the files that we
    # have in POTFILES.* actually exist.
    is_ok &= check_exists_in_tree(root, file_entries)

    # Now the other direction -- check that all sources
    # that should be included in POTFILES, are included.
    is_ok &= check_exists_in_potfiles(root, file_entries)

    return is_ok


if __name__ == "__main__":
    if xgettext is None:
        raise Exception("xgettext is missing")

    out_msg = "POTFILES consistency check: %s"
    if check_potfiles():
        print(out_msg % "ok")
    else:
        print(out_msg % "failed", file=sys.stderr)
        sys.exit(1)



#!/usr/bin/env python
# SPDX-License-Identifier: LGPL-2.1-or-later

import xml.etree.ElementTree as ET
import argparse
import re

###############################################################################


def strip_new_lines(desc_string):
    desc_string = re.sub(r"\n\s*", r" ", desc_string)
    return desc_string


def escape_quotes(desc_string):
    desc_string = re.sub(r"\"", r"\"", desc_string)
    return desc_string


###############################################################################


def main(output_path_str, xml_path_str):
    xml = ET.parse(xml_path_str).getroot()

    doc = ""
    doc += "/* Generated file. Do not edit. */\n\n"

    for setting in xml:
        name_set_upper = setting.attrib["name_upper"]

        for property in setting:
            name_prop_upper = property.attrib["name_upper"]
            desc_string = ""
            desc = property.find("description")

            if desc is not None:
                desc_string = desc.text

            desc_string = strip_new_lines(desc_string)
            desc_string = escape_quotes(desc_string)
            desc_string = desc_string.lstrip(" ")

            doc += "#define DESCRIBE_DOC_NM_SETTING_"
            doc += name_set_upper
            doc += "_"
            doc += name_prop_upper
            doc += ' N_("'
            doc += desc_string
            doc += '")\n'

    file = open(output_path_str, "w")
    file.write(doc)
    file.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output",
        metavar="PATH",
        help="specify output file",
        required=True,
    )
    parser.add_argument(
        "--xml",
        metavar="PATH",
        help="specify input file",
        required=True,
    )

    args = parser.parse_args()
    main(args.output, args.xml)



#!/usr/bin/env python
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# Copyright (C) 2009 - 2017 Red Hat, Inc.
#
from __future__ import print_function, unicode_literals
import xml.etree.ElementTree as ET
import argparse
import gi
import re

gi.require_version("NM", "1.0")
from gi.repository import NM, GObject

dbus_type_name_map = {
    "b": "boolean",
    "s": "string",
    "i": "int32",
    "u": "uint32",
    "t": "uint64",
    "x": "int64",
    "y": "byte",
    "as": "array of string",
    "au": "array of uint32",
    "ay": "byte array",
    "a{ss}": "dict of string to string",
    "a{sv}": "vardict",
    "aa{sv}": "array of vardict",
    "aau": "array of array of uint32",
    "aay": "array of byte array",
    "a(ayuay)": "array of legacy IPv6 address struct",
    "a(ayuayu)": "array of legacy IPv6 route struct",
}

ns_map = {
    "c": "http://www.gtk.org/introspection/c/1.0",
    "gi": "http://www.gtk.org/introspection/core/1.0",
    "glib": "http://www.gtk.org/introspection/glib/1.0",
}
identifier_key = "{%s}identifier" % ns_map["c"]
nick_key = "{%s}nick" % ns_map["glib"]
symbol_prefix_key = "{%s}symbol-prefix" % ns_map["c"]
nick_key = "{%s}nick" % ns_map["glib"]

constants = {
    "TRUE": "TRUE",
    "FALSE": "FALSE",
    "G_MAXUINT32": "G_MAXUINT32",
    "NULL": "NULL",
}
setting_names = {}


def get_setting_name_define(setting):
    n = setting.attrib[symbol_prefix_key]
    if n and n.startswith("setting_"):
        return n[8:].upper()
    raise Exception('Unexpected symbol_prefix_key "%s"' % (n))


def init_constants(girxml, settings, output_target):
    for const in girxml.findall("./gi:namespace/gi:constant", ns_map):
        cname = const.attrib["{%s}type" % ns_map["c"]]
        value = const.attrib["value"]
        if const.find('./gi:type[@name="utf8"]', ns_map) is not None:
            value = '"%s"' % value
        constants[cname] = value

    for enum in girxml.findall("./gi:namespace/gi:enumeration", ns_map):
        for enumval in enum.findall("./gi:member", ns_map):
            cname = enumval.attrib[identifier_key]
            nick = enumval.attrib.get(nick_key, cname)
            if output_target == "nmcli":
                value = '"%s" (%s)' % (nick, enumval.attrib["value"])
            elif output_target is not None:
                value = "%s (%s)" % (enumval.attrib["value"], nick)
            else:
                value = "%s (%s)" % (cname, enumval.attrib["value"])
            constants[cname] = value

    for enum in girxml.findall("./gi:namespace/gi:bitfield", ns_map):
        for enumval in enum.findall("./gi:member", ns_map):
            cname = enumval.attrib[identifier_key]
            nick = enumval.attrib.get(nick_key, cname)
            if output_target == "nmcli":
                value = '"%s" (0x%x)' % (nick, int(enumval.attrib["value"]))
            elif output_target is not None:
                value = "0x%x (%s)" % (int(enumval.attrib["value"]), nick)
            else:
                value = "%s (0x%x)" % (cname, int(enumval.attrib["value"]))
            constants[cname] = value

    for setting in settings:
        setting_type_name = "NM" + setting.attrib["name"]
        setting_name_symbol = (
            "NM_SETTING_" + get_setting_name_define(setting) + "_SETTING_NAME"
        )
        if setting_name_symbol in constants:
            setting_name = constants[setting_name_symbol]
            setting_names[setting_type_name] = setting_name


def get_prop_type(setting, pspec):
    dbus_type = setting.get_dbus_property_type(pspec.name).dup_string()
    prop_type = dbus_type_name_map[dbus_type]

    if GObject.type_is_a(pspec.value_type, GObject.TYPE_ENUM) or GObject.type_is_a(
        pspec.value_type, GObject.TYPE_FLAGS
    ):
        prop_type = "%s (%s)" % (pspec.value_type.name, prop_type)

    return prop_type


def remove_prefix(line, prefix):
    return line[len(prefix) :] if line.startswith(prefix) else line


def format_docs(doc_xml):
    doc = doc_xml.text

    # split docs into lines
    lines = re.split("\n", doc)
    # strip leading *char and strip white spaces
    lines = [remove_prefix(l, "*").strip() for l in lines]
    doc = ""
    for l in lines:
        if l:
            doc += l + " "
        else:
            doc = doc.strip(" ") + "\n\n"

    doc = doc.strip("\n ")

    # Expand constants
    doc = re.sub(r"%([^%]\w*)", lambda match: constants[match.group(1)], doc)

    # #NMSettingWired:mac-address -> "mac-address"
    doc = re.sub(r"#[A-Za-z0-9_]*:([A-Za-z0-9_-]*)", r'"\1"', doc)

    # #NMSettingWired setting -> "802-3-ethernet" setting
    doc = re.sub(
        r"#([A-Z]\w*) setting",
        lambda match: setting_names[match.group(1)] + " setting",
        doc,
    )

    # remaining gtk-doc cleanup
    doc = doc.replace("%%", "%")
    doc = doc.replace("<!-- -->", "")
    doc = re.sub(r" Element-.ype:.*", "", doc)
    doc = re.sub(r"#([A-Z]\w*)", r"\1", doc)

    # Remove sentences that refer to functions
    if "FindProxyForURL()" in doc:
        # FIXME: this would break the description for "proxy.pac-script"
        # Work around. But really the entire approach here is flawed
        # and needs improvement.
        pass
    else:
        doc = re.sub(r"\.\s+[^.]*\w\(\)[^.]*\.", r".", doc)

    return doc


def get_docs(propxml):
    doc_xml = propxml.find("gi:doc", ns_map)
    if doc_xml is None:
        return None
    else:
        return format_docs(doc_xml)


def get_default_value(setting, pspec, propxml):
    default_value = setting.get_property(pspec.name.replace("-", "_"))
    if default_value is None:
        return default_value

    value_type = get_prop_type(setting, pspec)
    if value_type == "string" and default_value != "" and pspec.name != "name":
        default_value = '"%s"' % default_value
    elif value_type == "boolean":
        default_value = str(default_value).upper()
    elif value_type == "byte array":
        default_value = "[]"
    elif str(default_value).startswith("<"):
        default_value = None
    elif str(default_value).startswith("["):
        default_value = None

    return default_value


def settings_sort_key(x):
    x_prefix = x.attrib["{%s}symbol-prefix" % ns_map["c"]]
    # always sort NMSettingConnection first
    return (x_prefix != "setting_connection", x_prefix)


def create_desc_docbook(desc_docbook, description):
    lines = re.split("\n", description)

    paragraph = ET.SubElement(
        desc_docbook,
        "para",
    )

    for l in lines:
        if not l:
            # A blank line. This starts a new paragraph
            paragraph = ET.SubElement(desc_docbook, "para")
            continue
        paragraph.text = l


def main(gir_path_str, output_path_str, output_target):
    girxml = ET.parse(gir_path_str).getroot()

    basexml = girxml.find('./gi:namespace/gi:class[@name="Setting"]', ns_map)
    settings = girxml.findall('./gi:namespace/gi:class[@parent="Setting"]', ns_map)
    # Hack. Need a better way to do this
    ipxml = girxml.find('./gi:namespace/gi:class[@name="SettingIPConfig"]', ns_map)
    settings.extend(
        girxml.findall('./gi:namespace/gi:class[@parent="SettingIPConfig"]', ns_map)
    )
    settings = sorted(settings, key=settings_sort_key)

    init_constants(girxml, settings, output_target)

    nm_settings_docs_element = ET.Element("nm-setting-docs")
    docs_gir = ET.ElementTree(nm_settings_docs_element)

    for settingxml in settings:
        if "abstract" in settingxml.attrib:
            continue

        new_func = NM.__getattr__(settingxml.attrib["name"])
        setting = new_func()

        class_desc = get_docs(settingxml)
        if class_desc is None:
            raise Exception(
                "%s needs a gtk-doc block with one-line description"
                % setting.props.name
            )
        setting_element = ET.SubElement(
            nm_settings_docs_element,
            "setting",
            attrib={
                "name": setting.props.name,
                "description": class_desc,
                "name_upper": get_setting_name_define(settingxml),
            },
        )

        setting_properties = {
            prop.name: prop
            for prop in GObject.list_properties(setting)
            if prop.name != "name"
        }

        for prop in sorted(setting_properties):
            pspec = setting_properties[prop]

            propxml = settingxml.find('./gi:property[@name="%s"]' % pspec.name, ns_map)
            if propxml is None:
                propxml = basexml.find('./gi:property[@name="%s"]' % pspec.name, ns_map)
            if propxml is None:
                propxml = ipxml.find('./gi:property[@name="%s"]' % pspec.name, ns_map)

            value_type = get_prop_type(setting, pspec)
            value_desc = get_docs(propxml)
            default_value = get_default_value(setting, pspec, propxml)

            if "deprecated" in propxml.attrib:
                deprecated = True
                deprecated_since = propxml.attrib["deprecated-version"]
                deprecated_desc = format_docs(propxml.find("gi:doc-deprecated", ns_map))
            else:
                deprecated = False

            prop_upper = prop.upper().replace("-", "_")

            if value_desc is None:
                raise Exception(
                    "%s.%s needs a documentation description"
                    % (setting.props.name, prop)
                )

            property_attributes = {
                "name": prop,
                "name_upper": prop_upper,
                "type": value_type,
            }

            if default_value is not None:
                property_attributes["default"] = str(default_value)

            property_element = ET.SubElement(
                setting_element,
                "property",
                attrib=property_attributes,
            )

            ET.SubElement(
                property_element,
                "description",
            ).text = value_desc

            if value_desc:
                description_docbook = ET.SubElement(
                    property_element,
                    "description-docbook",
                )

                create_desc_docbook(description_docbook, value_desc)

            if deprecated:
                ET.SubElement(
                    property_element,
                    "deprecated",
                    attrib={
                        "since": deprecated_since,
                    },
                ).text = deprecated_desc

                # The text should only be one line. Otherwise, our simple "<deprecated>" element
                # cannot be rendered nicely.
                assert re.split("\n", deprecated_desc) == [deprecated_desc]

    docs_gir.write(
        output_path_str,
        xml_declaration=True,
        encoding="utf-8",
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-g",
        "--gir",
        metavar="FILE",
        help="NM-1.0.gir file",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        help="output file",
        required=True,
    )
    parser.add_argument(
        "-t",
        "--target",
        choices=["nmcli", "dbus", "keyfile", "ifcfg-rh"],
        help="target where the output will be used (i.e. nmcli)",
    )

    args = parser.parse_args()

    main(args.gir, args.output, args.target)



#!/usr/bin/env python
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2014 Red Hat, Inc.
#

# This example lists all devices, both real and placeholder ones

import gi

gi.require_version("NM", "1.0")
from gi.repository import NM

if __name__ == "__main__":
    client = NM.Client.new(None)
    devices = client.get_all_devices()

    print("Real devices")
    print("------------")
    for d in devices:
        if d.is_real():
            print(
                "%s (%s): %s"
                % (
                    d.get_iface(),
                    d.get_type_description(),
                    d.get_state(),
                )
            )

    print("\nUnrealized/placeholder devices")
    print("------------------------------")
    for d in devices:
        if not d.is_real():
            print(
                "%s (%s): %s"
                % (
                    d.get_iface(),
                    d.get_type_description(),
                    d.get_state(),
                )
            )



#!/usr/bin/env python
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2013 Red Hat, Inc.
#

import locale
import math
import os
import re
import time

import gi

gi.require_version("NM", "1.0")
from gi.repository import NM, GLib, Gio

SCAN_THRESHOLD_MSEC = 10000

#
# This example lists Wi-Fi access points NetworkManager scanned on Wi-Fi devices.
# It calls libnm functions using GObject introspection.
#
# Note the second line of the file: coding=utf-8
# It is necessary because we use unicode characters and python would produce
# an error without it: http://www.python.org/dev/peps/pep-0263/
#

NM80211Mode = getattr(NM, "80211Mode")
NM80211ApFlags = getattr(NM, "80211ApFlags")
NM80211ApSecurityFlags = getattr(NM, "80211ApSecurityFlags")

main_loop = GLib.MainLoop()


def gflags_to_str(flags_type, value):
    if value == 0:
        return "none"
    str = ""
    for n in sorted(dir(flags_type)):
        if not re.search("^[A-Z0-9_]+$", n):
            continue
        flag_value = getattr(flags_type, n)
        if value & flag_value:
            value &= ~flag_value
            str += " " + n
            if value == 0:
                break
    if value:
        str += " (0x%0x)" % (value,)
    return str.lstrip()


def genum_to_str(enum_type, value):
    for n in sorted(dir(enum_type)):
        if not re.search("^[A-Z0-9_]+$", n):
            continue
        enum_value = getattr(enum_type, n)
        if value == enum_value:
            return n
    return "(%d" % (value,)


def ap_security_flags_to_security(flags, wpa_flags, rsn_flags):
    str = ""
    if (flags & NM80211ApFlags.PRIVACY) and (wpa_flags == 0) and (rsn_flags == 0):
        str = str + " WEP"
    if wpa_flags != 0:
        str = str + " WPA1"
    if rsn_flags != 0:
        str = str + " WPA2"
    if (wpa_flags & NM80211ApSecurityFlags.KEY_MGMT_802_1X) or (
        rsn_flags & NM80211ApSecurityFlags.KEY_MGMT_802_1X
    ):
        str = str + " 802.1X"
    return str.lstrip()


def ap_get_ssid(ap):
    if ap is None:
        return "not connected"
    ssid = ap.get_ssid()
    if ssid is None:
        return "no ssid"
    return '"%s"' % (NM.utils_ssid_to_utf8(ssid.get_data()),)


def print_device_info(device):
    if device.get_client() is None:
        last_scan = "device disappeared"
    else:
        t = device.get_last_scan()
        if t == 0:
            last_scan = "no scan completed"
        else:
            t = (NM.utils_get_timestamp_msec() - t) / 1000.0
            last_scan = "%0.2f sec ago" % (t,)
            if device_needs_scan(device):
                last_scan += " (stale)"

    ap = device.get_active_access_point()
    if ap is None:
        active_ap = "none"
    else:
        active_ap = "%s (%s)" % (ap_get_ssid(ap), ap.get_path())

    print("Device:     %s" % (device.get_iface(),))
    print("D-Bus path: %s" % (NM.Object.get_path(device),))
    print("Driver:     %s" % (device.get_driver(),))
    print("Active AP:  %s" % (active_ap,))
    print("Last scan:  %s" % (last_scan,))


def print_ap_info(ap):
    strength = ap.get_strength()
    frequency = ap.get_frequency()
    flags = ap.get_flags()
    wpa_flags = ap.get_wpa_flags()
    rsn_flags = ap.get_rsn_flags()

    t = ap.get_last_seen()
    if t < 0:
        last_seen = "never"
    else:
        t = time.clock_gettime(time.CLOCK_BOOTTIME) - t
        last_seen = "%s sec ago" % (math.ceil(t),)

    print("  - D-Bus path: %s" % (ap.get_path(),))
    print("    SSID:       %s" % (ap_get_ssid(ap),))
    print("    BSSID:      %s" % (ap.get_bssid(),))
    print("    Last seen:  %s" % (last_seen,))
    print("    Frequency:  %s" % (frequency,))
    print("    Channel:    %s" % (NM.utils_wifi_freq_to_channel(frequency),))
    print("    Mode:       %s" % (genum_to_str(NM80211Mode, ap.get_mode()),))
    print("    Flags:      %s" % (gflags_to_str(NM80211ApFlags, flags),))
    print("    WPA flags:  %s" % (gflags_to_str(NM80211ApSecurityFlags, wpa_flags),))
    print("    RSN flags:  %s" % (gflags_to_str(NM80211ApSecurityFlags, rsn_flags),))
    print(
        "    Security:   %s"
        % (ap_security_flags_to_security(flags, wpa_flags, rsn_flags),)
    )
    print(
        "    Strength:   %s%% : %s"
        % (
            strength,
            NM.utils_wifi_strength_bars(strength),
        )
    )


def device_needs_scan(device):
    if device.get_client() is None:
        # the device got deleted. We can forget about it.
        return False
    t = device.get_last_scan()
    return t == 0 or t < NM.utils_get_timestamp_msec() - SCAN_THRESHOLD_MSEC


def device_ensure_scanned(device):
    if os.getenv("NO_SCAN") == "1":
        return

    if not device_needs_scan(device):
        return

    # kick off a new scan.
    device.request_scan_async(None)

    def cb():
        main_loop.quit()

    timeout_source = GLib.timeout_source_new(10 * 1000)
    timeout_source.set_callback(cb)
    timeout_source.attach(main_loop.get_context())

    def cb(device, prop):
        if not device_needs_scan(device):
            main_loop.quit()

    device.connect("notify", cb)

    main_loop.run()

    timeout_source.destroy()


def main():
    # Python apparently doesn't call setlocale() on its own? We have to call this or else
    # NM.utils_wifi_strength_bars() will think the locale is ASCII-only, and return the
    # fallback characters rather than the unicode bars
    locale.setlocale(locale.LC_ALL, "")

    nmc = NM.Client.new(None)
    devs = nmc.get_devices()

    is_first = True
    for device in devs:
        if device.get_device_type() != NM.DeviceType.WIFI:
            continue

        if not is_first:
            print("")
        else:
            is_first = False

        device_ensure_scanned(device)
        print_device_info(device)
        for ap in device.get_access_points():
            print_ap_info(ap)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2017, 2020 Red Hat, Inc.
#

#
# set and show OVS external-ids and other-config for a connection:
#

import collections
import os
import pprint
import re
import subprocess
import sys

import gi

gi.require_version("NM", "1.0")
from gi.repository import GLib, NM

###############################################################################

MODE_GET = "get"
MODE_SET = "set"
MODE_APPLY = "apply"


def memoize0(f):
    result = []

    def helper():
        if len(result) == 0:
            result.append(f())
        return result[0]

    return helper


def memoize(f):
    memo = {}

    def helper(x):
        if x not in memo:
            memo[x] = f(x)
        return memo[x]

    return helper


def pr(v):
    pprint.pprint(v, indent=4, depth=5, width=60)


@memoize0
def is_libnm_debug():
    return os.getenv("LIBNM_CLIENT_DEBUG") is not None


@memoize0
def can_sudo():
    try:
        return (
            subprocess.run(
                ["sudo", "-n", "true"],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            ).returncode
            == 0
        )
    except Exception:
        return False


def _print(msg=""):
    if is_libnm_debug():
        # we want to use the same logging mechanism as libnm's debug
        # logging with "LIBNM_CLIENT_DEBUG=trace,stdout".
        NM.utils_print(0, msg + "\n")
        return
    print(msg)


def mainloop_run(timeout_msec=0, mainloop=None):
    if mainloop is None:
        mainloop = GLib.MainLoop()

    timeout_id = None
    timeout_reached = []

    if timeout_msec > 0:

        def _timeout_cb(unused):
            # it can happen that the caller already quit the mainloop
            # otherwise. In that case, we don't want to signal a timeout.
            if mainloop.is_running():
                timeout_reached.append(1)
                mainloop.quit()
            return True

        timeout_id = GLib.timeout_add(timeout_msec, _timeout_cb, None)

    mainloop.run()
    if timeout_id:
        GLib.source_remove(timeout_id)
    return not timeout_reached


###############################################################################


def connection_update2(remote_connection, connection):
    mainloop = GLib.MainLoop()
    result_error = []

    def cb(c, result):
        try:
            c.update2_finish(result)
        except Exception as e:
            result_error.append(e)
        mainloop.quit()

    remote_connection.update2(
        connection.to_dbus(NM.ConnectionSerializationFlags.ALL),
        NM.SettingsUpdate2Flags.NO_REAPPLY,
        None,
        None,
        cb,
    )

    mainloop_run(mainloop=mainloop)

    if result_error:
        raise result_error[0]


def device_get_applied_connection(device):
    mainloop = GLib.MainLoop()
    rr = []

    def cb(c, result):
        try:
            con, version_id = c.get_applied_connection_finish(result)
        except Exception as e:
            rr.append(e)
        else:
            rr.append(con)
            rr.append(version_id)
        mainloop.quit()

    device.get_applied_connection_async(0, None, cb)

    mainloop_run(mainloop=mainloop)

    if len(rr) == 1:
        raise rr[0]
    return rr[0], rr[1]


def device_reapply(device, connection, version_id):
    mainloop = GLib.MainLoop()
    result_error = []

    def cb(d, result):
        try:
            d.reapply_finish(result)
        except Exception as e:
            result_error.append(e)
        mainloop.quit()

    device.reapply_async(connection, version_id, 0, None, cb)

    mainloop_run(mainloop=mainloop)

    if len(result_error) == 1:
        raise result_error[0]


def ovs_print_config(prefix):
    if not can_sudo():
        _print(prefix + ": not running as root and cannot call ovs-vsctl")
        return

    cmds = [["ovs-vsctl", "show"]]
    for typ in ["Bridge", "Port", "Interface"]:
        cmds += [["ovs-vsctl", "--columns=name,external-ids,other-config", "list", typ]]

    out = ""
    for cmd in cmds:
        p = subprocess.run(cmd, stdout=subprocess.PIPE, check=True)
        out += p.stdout.decode("utf-8") + "\n"
    out = "\n".join([prefix + s for s in out.split("\n")])
    _print(out)


###############################################################################


def usage():
    _print("%s g[et]   PROFILE [ GETTER ]" % (sys.argv[0]))
    _print("%s s[et]   PROFILE   SETTER   [--test]" % (sys.argv[0]))
    _print("%s a[pply] DEVICE    SETTER   [--test]" % (sys.argv[0]))
    _print(
        "   PROFILE :=  [id | uuid | type] STRING  |  [ ~id | ~type ] REGEX_STRING  |  STRING"
    )
    _print("   DEVICE :=  [iface] STRING")
    _print("   GETTER  := ( KEY | ~REGEX_KEY )  [... GETTER]")
    _print("   SETTER  := ( +[e:|o:] | -[e:|o:] | -KEY | [+]KEY VALUE ) [... SETTER]")
    _print("")
    _print(
        'Prefix KEY with "e:" or "o:" to set external-ids or other-config ("e:" is the default)'
    )


def die(msg, show_usage=False):
    _print("FAILED: %s" % (msg))
    if show_usage:
        usage()
    sys.exit(1)


def die_usage(msg):
    die(msg, show_usage=True)


DataTypeTuple = collections.namedtuple(
    "DataTypeTuple", ["short", "name", "setting_type", "property_name"]
)

DataTypeE = DataTypeTuple(
    "external-ids",
    "ovs-external-ids",
    NM.SettingOvsExternalIDs,
    NM.SETTING_OVS_EXTERNAL_IDS_DATA,
)

o = getattr(NM, "SettingOvsOtherConfig", None)
DataTypeO = DataTypeTuple(
    "other-config",
    "ovs-other-config",
    o,
    NM.SETTING_OVS_OTHER_CONFIG_DATA if o else None,
)


def parse_args(argv):
    args = {
        "mode": MODE_GET,
        "select_arg": None,
        "ids_arg": [],
        "do_test": False,
    }
    i = 1
    while i < len(argv):
        a = argv[i]

        if i == 1:
            if a in ["s", "set"]:
                args["mode"] = MODE_SET
            elif a in ["g", "get"]:
                args["mode"] = MODE_GET
            elif a in ["a", "apply"]:
                args["mode"] = MODE_APPLY
            else:
                die_usage("unexpected mode argument '%s'" % (a))
            i += 1
            continue

        if a == "--test":
            args["do_test"] = True
            i += 1
            continue

        if args["select_arg"] is None:
            if args["mode"] == MODE_APPLY:
                possible_selects = ["iface"]
            else:
                possible_selects = ["id", "~id", "uuid", "type", "~type"]

            if a in possible_selects:
                if i + 1 >= len(argv):
                    die_usage("'%s' requires an argument'" % (a))
                args["select_arg"] = (a, argv[i + 1])
                i += 2
                continue

            if a == "*":
                a = None
            args["select_arg"] = ("*", a)
            i += 1
            continue

        if args["mode"] == MODE_GET:
            args["ids_arg"].append(a)
            i += 1
            continue

        if not a:
            die_usage(
                "argument should specify a external-id/other-config but is empty string"
            )

        if a[0] == "-":
            v = (a, None)
            i += 1
        elif a in ["+", "+o:", "+e:"]:
            v = (a, None)
            i += 1
        else:
            if a[0] != "+":
                a = "+" + a
            if i + 1 >= len(argv):
                die_usage("'%s' requires an argument'" % (a))
            v = (a, argv[i + 1])
            i += 2

        args["ids_arg"].append(v)

    if args["mode"] == MODE_SET:
        if not args["ids_arg"]:
            die_usage("Requires one or more external-ids/other-config to set or delete")

    return args


def device_to_str(device, show_type=False):
    if show_type:
        return "%s (%s)" % (device.get_iface(), device.get_type_desc())
    return "%s" % (device.get_iface(),)


def connection_to_str(connection, show_type=False):
    if show_type:
        return "%s (%s, %s)" % (
            connection.get_id(),
            connection.get_uuid(),
            connection.get_connection_type(),
        )
    return "%s (%s)" % (connection.get_id(), connection.get_uuid())


def devices_filter(devices, select_arg):
    devices = list(sorted(devices, key=device_to_str))
    if not select_arg:
        return devices
    l = []
    f = select_arg
    for d in devices:
        if f[0] == "iface":
            if f[1] == d.get_iface():
                l.append(d)
        else:
            assert f[0] == "*"
            if f[1] is None:
                l.append(d)
            else:
                if f[1] in [d.get_iface()]:
                    l.append(d)
    return l


def connections_filter(connections, select_arg):
    connections = list(sorted(connections, key=connection_to_str))
    if not select_arg:
        return connections
    l = []
    f = select_arg
    for c in connections:
        if f[0] == "id":
            if f[1] == c.get_id():
                l.append(c)
        elif f[0] == "~id":
            if re.search(f[1], c.get_id()):
                l.append(c)
        elif f[0] == "uuid":
            if f[1] == c.get_uuid():
                l.append(c)
        elif f[0] == "type":
            if f[1] == c.get_connection_type():
                l.append(c)
        elif f[0] == "~type":
            if re.search(f[1], c.get_connection_type()):
                l.append(c)
        else:
            assert f[0] == "*"
            if f[1] is None:
                l.append(c)
            else:
                if f[1] in [c.get_uuid(), c.get_id()]:
                    l.append(c)
    return l


def ids_select(ids, mode, ids_arg):
    ids = list(ids)
    if not ids_arg:
        return (ids, [])

    keys = set()
    requested = []
    for d in ids_arg:
        if mode == MODE_GET:
            if d[0] == "~":
                r = re.compile(d[1:])
                keys.update([k for k in ids if r.search(k)])
            else:
                keys.update([k for k in ids if k == d])
                if d not in requested:
                    requested.append(d)
        else:
            assert mode in [MODE_SET, MODE_APPLY]
            d2 = d[0]
            assert d2[0] in ["-", "+"]
            d3 = d2[1:]
            if d3 in ids:
                keys.add(d3)
    return (list([k for k in ids if k in keys]), requested)


def connection_print(connection, mode, ids_arg, dbus_path, prefix=""):
    def _num_str(connection, data_type):
        if data_type.setting_type is None:
            return "n/a"
        sett = connection.get_setting(data_type.setting_type)
        if sett is None:
            return "none"
        all_ids = list(sett.get_data_keys())
        return "%s" % (len(all_ids),)

    _print(
        "%s%s [e:%s, o:%s]"
        % (
            prefix,
            connection_to_str(connection, show_type=True),
            _num_str(connection, DataTypeE),
            _num_str(connection, DataTypeO),
        )
    )
    if dbus_path:
        _print("%s   %s" % (prefix, dbus_path))

    for data_type in [DataTypeE, DataTypeO]:
        if data_type.setting_type is None:
            continue

        sett = connection.get_setting(data_type.setting_type)
        if sett is None:
            continue

        all_ids = list(sett.get_data_keys())
        keys, requested = ids_select(all_ids, mode, ids_arg)

        dd = sett.get_property(data_type.property_name)
        for k in keys:
            v = sett.get_data(k)
            assert v is not None
            assert v == dd.get(k, None)
            _print('%s   %s: "%s" = "%s"' % (prefix, data_type.short, k, v))
        for k in requested:
            _print('%s   %s: "%s" = <unset>' % (prefix, data_type.short, k))


def sett_update(connection, ids_arg):
    for d in ids_arg:
        op = d[0][0]
        key = d[0][1:]
        val = d[1]

        if key == "o" or key.startswith("o:"):
            data_type = DataTypeO
            key = key[2:]
        elif key == "e" or key.startswith("e:"):
            data_type = DataTypeE
            key = key[2:]
        else:
            data_type = DataTypeE

        if data_type.setting_type is None:
            raise Exception(
                "%s is not supported by this version of libnm" % (data_type.name,)
            )

        sett = connection.get_setting(data_type.setting_type)

        oldval = None
        if sett is not None:
            oldval = sett.get_data(key)

        if op == "-":
            assert val is None
            if key == "":
                if sett is None:
                    _print(
                        " DEL: setting (%s group was not present)" % (data_type.name,)
                    )
                else:
                    connection.remove_setting(data_type.setting_type)
                    sett = None
                    _print(" DEL: setting (%s)" % (data_type.name,))
                continue

            if sett is None:
                _print(' DEL: "%s" (%s group was not present)' % (key, data_type.name))
                continue
            if oldval is None:
                _print(' DEL: "%s" (id was unset)' % (key))
                continue
            _print(' DEL: "%s" (id was set to"%s")' % (key, oldval))
            sett.set_data(key, None)
            continue

        if key == "":
            assert val is None
            if sett is None:
                sett = data_type.setting_type.new()
                connection.add_setting(sett)
                _print(" SET: setting (%s group was added)" % (data_type.name,))
                continue

            _print(" SET: setting (%s group was present)" % (data_type.name,))
            continue

        assert val is not None

        if sett is None:
            sett = data_type.setting_type.new()
            connection.add_setting(sett)
            _print(
                ' SET: "%s" = "%s" (%s group was not present)'
                % (key, val, data_type.name)
            )
        elif oldval is None:
            _print(' SET: "%s" = "%s" (new)' % (key, val))
        elif oldval != val:
            _print(' SET: "%s" = "%s" (was "%s")' % (key, val, oldval))
        else:
            _print(' SET: "%s" = "%s" (unchanged)' % (key, val))
        sett.set_data(key, val)


def do_get(connections, ids_arg):
    first_line = True
    for c in connections:
        if first_line:
            first_line = False
        else:
            _print()
        connection_print(c, MODE_GET, ids_arg, dbus_path=c.get_path())


def do_set(nmc, connection, ids_arg, do_test):
    remote_connection = connection
    connection = NM.SimpleConnection.new_clone(remote_connection)

    connection_print(
        connection, MODE_SET, [], remote_connection.get_path(), prefix="BEFORE: "
    )
    _print()

    sett_update(connection, ids_arg)

    if do_test:
        _print()
        _print("Only show. Run without --test to set")
        return

    try:
        connection_update2(remote_connection, connection)
    except Exception as e:
        _print()
        _print("FAILURE to commit connection: %s" % (e))
        return

    # NMClient received the completion of Update2() call. It also received
    # a property changed signal that the profile changed, and it is about
    # to fetch the new value. However, that value is not yet here.
    #
    # libnm should provide a better API for this. For example, not signal
    # completion of update2() until the profile was refetched. Or, indicate
    # that the settings are dirty, so we would know how long to wait.
    #
    # Add an ugly workaround here and wait a bit.
    _print()
    _print("WORKAROUND: wait for connection to change")
    mainloop_run(timeout_msec=500)

    if remote_connection is not nmc.get_object_by_path(remote_connection.get_path()):
        _print()
        _print(
            "Connection %s no longer exists after commit"
            % (remote_connection.get_path())
        )
        return

    _print()
    connection_print(
        remote_connection, MODE_SET, [], remote_connection.get_path(), prefix="AFTER:  "
    )

    _print()
    if remote_connection.compare(connection, NM.SettingCompareFlags.EXACT):
        _print("resulting connection is as expected")
    else:
        _print("WARNING: resulting connection is not as expected")


def do_apply(nmc, device, ids_arg, do_test):
    try:
        connection_orig, version_id = device_get_applied_connection(device)
    except Exception as e:
        _print(
            'failure to get applied connection for %s: %s"' % (device_to_str(device), e)
        )
        die("The device does not seem active? Nothing to reapply")

    _print(
        "REAPPLY device %s (%s) with connection %s (version-id = %s)"
        % (
            device_to_str(device),
            NM.Object.get_path(device),
            connection_to_str(connection_orig),
            version_id,
        )
    )
    _print()

    ovs_print_config("BEFORE-OVS-VSCTL: ")
    _print()

    connection = NM.SimpleConnection.new_clone(connection_orig)

    connection_print(connection, MODE_APPLY, [], device.get_path(), prefix="BEFORE: ")
    _print()

    sett_update(connection, ids_arg)

    if do_test:
        _print()
        _print("Only show. Run without --test to set")
        return

    _print()
    _print("reapply...")

    try:
        device_reapply(device, connection, version_id)
    except Exception as e:
        _print()
        _print("FAILURE to commit connection: %s" % (e))
        return

    try:
        connection_after, version_id = device_get_applied_connection(device)
    except Exception as e:
        _print(
            'failure to get applied connection after reapply for device %s: %s"'
            % (device_to_str(device), e)
        )
        die("FAILURE to get applied connection after reapply")

    _print()
    connection_print(
        connection_after, MODE_APPLY, [], device.get_path(), prefix="AFTER: "
    )
    _print()

    ovs_print_config("AFTER-OVS-VSCTL: ")


###############################################################################

if __name__ == "__main__":
    args = parse_args(sys.argv)

    nmc = NM.Client.new(None)

    if args["mode"] == MODE_APPLY:
        devices = devices_filter(nmc.get_devices(), args["select_arg"])

        if len(devices) != 1:
            _print(
                "To apply the external-ids/other-config of a device, exactly one connection must be selected. Instead, %s devices matched ([%s])"
                % (len(devices), ", ".join([device_to_str(c) for c in devices]))
            )
            die_usage("Select unique device to apply")
        do_apply(nmc, devices[0], args["ids_arg"], do_test=args["do_test"])

    else:
        connections = connections_filter(nmc.get_connections(), args["select_arg"])

        if args["mode"] == MODE_SET:
            if len(connections) != 1:
                _print(
                    "To set the external-ids/other-config of a connection, exactly one connection must be selected via id|uuid. Instead, %s connection matched ([%s])"
                    % (
                        len(connections),
                        ", ".join([connection_to_str(c) for c in connections]),
                    )
                )
                die_usage("Select unique connection to set")
            do_set(nmc, connections[0], args["ids_arg"], do_test=args["do_test"])
        else:
            if len(connections) < 1:
                _print(
                    "No connection selected for printing the external ids/other-config"
                )
                die_usage("Select connection to get")
            do_get(connections, args["ids_arg"])



#!/usr/bin/env python
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2017 Red Hat, Inc.
#

import sys
import re

import gi

gi.require_version("NM", "1.0")
from gi.repository import GLib, NM

###############################################################################


def usage():
    print("Usage: %s [COMMAND [ARG]...]" % sys.argv[0])
    print("")
    print(" COMMANDS:  [show]")
    print(
        "            create TIMEOUT [--destroy-all|--delete-new-connections|--disconnect-new-devices|--allow-overlapping|DEV]..."
    )
    print("            destroy ['--last'|PATH|NUMBER[")
    print("            rollback ['--last'|PATH|NUMBER]")
    print("            adjust-rollback-timeout '--last'|PATH|NUMBER TIMEOUT")
    print("")
    print(" For destroy|rollback, when omitted then '--last' is the default.")
    sys.exit(1)


def show(c, ts=None):
    cr = c.get_created()
    rt = c.get_rollback_timeout()
    print("%s:" % c.get_path())
    print(
        "  created: %u%s"
        % (cr, "" if ts is None else (" (%s sec ago)" % ((ts - cr) / 1000.0)))
    )
    if rt == 0:
        print("  timeout: infinity")
    else:
        print(
            "  timeout: %u seconds%s"
            % (
                rt,
                (
                    ""
                    if ts is None
                    else (" (circa %s sec left)" % ((cr + (rt * 1000) - ts) / 1000.0))
                ),
            )
        )
    print(
        "  devices: %s"
        % (" ".join(sorted(map(lambda x: x.get_iface(), c.get_devices()))))
    )


def checkpoint_path_to_num(path):
    m = re.match(r"^/org/freedesktop/NetworkManager/Checkpoint/([1-9][0-9]*)$", path)
    if m:
        return int(m.group(1))
    raise Exception(f'Unexpected D-Bus path "{path}"for checkpoint')


def find_checkpoint(nmc, path):
    for c in nmc.get_checkpoints():
        if c.get_path() == path:
            return c
    return None


def find_checkpoint_last(nmc):
    return max(
        nmc.get_checkpoints(),
        key=lambda c: checkpoint_path_to_num(c.get_path()),
        default=None,
    )


def validate_path(path, nmc):
    if path == "--last":
        c = find_checkpoint_last(nmc)
        if c is None:
            sys.exit("Has no checkpoint")
        return c.get_path()

    try:
        num = int(path)
        path = f"/org/freedesktop/NetworkManager/Checkpoint/{num}"
    except Exception as e:
        pass

    if not path or path[0] != "/":
        sys.exit('Invalid checkpoint path "%s"' % (path))

    if nmc is not None:
        checkpoint = find_checkpoint(nmc, path)
        if checkpoint is None:
            print('WARNING: no checkpoint with path "%s" found' % (path))

    return path


def validate_path_from_argv(nmc):
    assert len(sys.argv) >= 2
    if len(sys.argv) == 2:
        path = "--last"
    elif len(sys.argv) > 3:
        sys.exit("Failed: invalid extra argument")
    else:
        path = sys.argv[2]

    return validate_path(path, nmc)


def do_create(nmc):
    flags = NM.CheckpointCreateFlags.NONE
    if len(sys.argv) < 3:
        sys.exit("Failed: missing argument timeout")

    timeout = int(sys.argv[2])
    devices = []
    for arg in sys.argv[3:]:
        if arg == "--destroy-all":
            flags |= NM.CheckpointCreateFlags.DESTROY_ALL
        elif arg == "--delete-new-connections":
            flags |= NM.CheckpointCreateFlags.DELETE_NEW_CONNECTIONS
        elif arg == "--disconnect-new-devices":
            flags |= NM.CheckpointCreateFlags.DISCONNECT_NEW_DEVICES
        elif arg == "--allow-overlapping":
            flags |= NM.CheckpointCreateFlags.ALLOW_OVERLAPPING
        else:
            d = nmc.get_device_by_iface(arg)
            if d is None:
                sys.exit("Unknown device %s" % arg)
            devices.append(d)

    def create_cb(nmc, result, data):
        try:
            checkpoint = nmc.checkpoint_create_finish(result)
            print("%s" % checkpoint.get_path())
        except Exception as e:
            sys.stderr.write("Failed: %s\n" % e.message)
        main_loop.quit()

    nmc.checkpoint_create(devices, timeout, flags, None, create_cb, None)


def do_destroy(nmc):
    path = validate_path_from_argv(nmc)

    def destroy_cb(nmc, result, data):
        try:
            if nmc.checkpoint_destroy_finish(result) == True:
                print("Success")
        except Exception as e:
            sys.stderr.write("Failed: %s\n" % e.message)
        main_loop.quit()

    nmc.checkpoint_destroy(path, None, destroy_cb, None)


def do_rollback(nmc):
    path = validate_path_from_argv(nmc)

    def rollback_cb(nmc, result, data):
        try:
            res = nmc.checkpoint_rollback_finish(result)
            for path in res:
                d = nmc.get_device_by_path(path)
                if d is None:
                    iface = path
                else:
                    iface = d.get_iface()
                print("%s => %s" % (iface, "OK" if res[path] == 0 else "ERROR"))
        except Exception as e:
            sys.stderr.write("Failed: %s\n" % e.message)
        main_loop.quit()

    nmc.checkpoint_rollback(path, None, rollback_cb, None)


def do_adjust_rollback_timeout(nmc):
    if len(sys.argv) < 3:
        sys.exit("Missing checkpoint path")
    if len(sys.argv) < 4:
        sys.exit("Missing timeout")
    try:
        add_timeout = int(sys.argv[3])
    except Exception:
        sys.exit("Invalid timeout")

    path = validate_path(sys.argv[2], nmc)

    def adjust_rollback_timeout_cb(nmc, result, data):
        try:
            nmc.checkpoint_adjust_rollback_timeout_finish(result)
            print("Success")
        except Exception as e:
            sys.stderr.write("Failed: %s\n" % e.message)
        main_loop.quit()

    nmc.checkpoint_adjust_rollback_timeout(
        path, add_timeout, None, adjust_rollback_timeout_cb, None
    )


def do_show(nmc):
    ts = NM.utils_get_timestamp_msec()
    for c in nmc.get_checkpoints():
        show(c, ts)


if __name__ == "__main__":
    nmc = NM.Client.new(None)
    main_loop = GLib.MainLoop()

    if len(sys.argv) < 2 or sys.argv[1] == "show":
        do_show(nmc)
        sys.exit(0)
    elif sys.argv[1] == "create":
        do_create(nmc)
    elif sys.argv[1] == "destroy":
        do_destroy(nmc)
    elif sys.argv[1] == "rollback":
        do_rollback(nmc)
    elif sys.argv[1] == "adjust-rollback-timeout":
        do_adjust_rollback_timeout(nmc)
    else:
        usage()

    main_loop.run()



#!/usr/bin/env python
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2014 Red Hat, Inc.
#

#
# This example imports a VPN connection, by loading the glib based
# VPN plugin.

import gi

gi.require_version("NM", "1.0")
from gi.repository import GLib, NM

import sys

if len(sys.argv) != 2:
    print("Expects one argument: the filename")
    sys.exit(1)
filename = sys.argv[1]

connection = None
for vpn_info in NM.VpnPluginInfo.list_load():
    print("TRY:  plugin %s" % (vpn_info.get_filename()))
    try:
        vpn_plugin = vpn_info.load_editor_plugin()
    except Exception as e:
        print("SKIP: cannot load plugin: %s" % (e))
        continue
    try:
        connection = vpn_plugin.import_(filename)
    except Exception as e:
        print("SKIP: failure to import %s" % (e))
        continue
    break

if connection is None:
    print('None of the VPN plugins was able to import "%s"' % (filename))
    sys.exit(1)

connection.normalize()

print(
    'connection imported from "%s" using plugin "%s" ("%s", %s)'
    % (filename, vpn_info.get_filename(), connection.get_id(), connection.get_uuid())
)

client = NM.Client.new(None)

main_loop = GLib.MainLoop()


def added_cb(client, result, data):
    try:
        client.add_connection_finish(result)
        print("The connection profile has been successfully added to NetworkManager.")
    except Exception as e:
        print("ERROR: failed to add connection: %s\n" % e)
    main_loop.quit()


client.add_connection_async(connection, True, None, added_cb, None)

main_loop.run()



#!/usr/bin/env python3

import datetime
import os
import sys

try:
    import yaml
except ImportError:
    print("Error: missing pyyaml. Install with `pip install pyyaml`.", file=sys.stderr)
    quit(code=1)


# These are the distros that we currently check, ordered by priority to be chosen as Tier 1
ci_distros = ("fedora", "centos", "debian", "ubuntu", "alpine")


def _parse_date(date_str) -> datetime.date:
    return datetime.datetime.strptime(date_str, "%Y-%m-%d").date()


def _is_supported(val_str, today) -> bool:
    val_str = val_str.lower()
    if val_str in ("yes", "true"):
        return True
    elif val_str in ("no", "false"):
        return False
    else:
        support_date = _parse_date(val_str)
        return today <= support_date


def _nm_version_is_newer(nm_ver, nm_ver_from):
    if nm_ver == "main":
        return nm_ver_from != "main"  # main is newer than anything except main itself
    elif nm_ver_from == "main":
        return False
    nm_ver = nm_ver.split(".")
    nm_ver_from = nm_ver_from.split(".")
    if int(nm_ver[0]) > int(nm_ver_from[0]):
        return True
    elif nm_ver[0] == nm_ver_from[0] and int(nm_ver[1]) > int(nm_ver_from[1]):
        return True
    return False


def _print_usage():
    print("Usage: distros_support.py [-a|--all] | <nm_version>")
    print(" -a|--all:   print NM versions still active in any distro")
    print(" nm_version: print all info and config.yml file of the specified NM version")


if len(sys.argv) == 2 and sys.argv[1] in ("-h", "--help", "help"):
    _print_usage()
    quit()
elif len(sys.argv) > 2:
    print("Error: wrong arguments.", file=sys.stderr)
    _print_usage()
    quit(code=1)

today = datetime.date.today()
with open(os.path.dirname(__file__) + "/distros-info.yml") as f:
    distros_info = yaml.load(f, Loader=yaml.BaseLoader)

# Warn about EOL'd distros to remove them
for distro, versions in distros_info.items():
    for info in versions:
        if _is_supported(info["support"], today):
            continue
        if "extended-support" in info and _is_supported(
            info["extended-support"], today
        ):
            continue
        print(
            f"Warn: {distro} {info['version']} reached EOL, consider deleting this entry",
            file=sys.stderr,
        )

# If --all is selected, print all active NM versions and return
if len(sys.argv) < 2 or sys.argv[1] in ("-a", "--all"):
    nm_versions = {}

    for distro, versions in distros_info.items():
        for info in versions:
            if not _is_supported(info["support"], today):
                continue
            nm_versions.setdefault(info["nm"], []).append(f"{distro} {info['version']}")

    for nm_ver, distros in sorted(nm_versions.items(), reverse=True):
        print("- NM {}: {}".format(nm_ver, ", ".join(distros)))

    quit()

# Otherwise, print all the info related to the specified NM version
nm_version = sys.argv[1]

# Print distros that uses this nm_version
print(f"# List of distros using NM {nm_version}")
print("---")
for distro, versions in distros_info.items():
    for info in versions:
        if nm_version == info["nm"] and _is_supported(info["support"], today):
            try:
                support_end_date = _parse_date(info["support"])
                print(
                    f"- {distro} {info['version']}, supported until {info['support']}"
                )
            except ValueError:
                print(f"- {distro} {info['version']}, supported")

# Collect info about what distros should be Tier 2 and 3
tier2 = {}
tier3 = {}
for distro, versions in distros_info.items():
    if distro not in ci_distros:
        continue
    for info in versions:
        if not _is_supported(info["support"], today):
            continue
        if nm_version == info["nm"]:
            tier2.setdefault(distro, []).append(info["version"])
        elif _nm_version_is_newer(nm_version, info["nm"]):
            tier3.setdefault(distro, []).append(info["version"])

# Select a Tier1 distro
tier1_distro, tier1_version = "", ""

for fed_ver_info in distros_info["fedora"]:
    # We prefer the Fedora version marked as tier1-default
    if fed_ver_info.get("tier1-default", False):
        for tier in (tier2, tier3):
            if fed_ver_info["version"] in tier.get("fedora", []):
                tier1_distro = "fedora"
                tier1_version = fed_ver_info["version"]
                tier["fedora"].remove(fed_ver_info["version"])
                if not tier["fedora"]:
                    del tier["fedora"]

for distro in ci_distros:
    if tier1_distro:
        break

    for tier in (tier2, tier3):
        if distro in tier:
            # Exception: we want to use fedora:latest instead of fedora:rawhide because
            # we don't want lot of build failures in Tier 1, which is run for every MR.
            # We just ignore fedora:rawhide for Tier 1.
            if distro == "fedora" and tier[distro][0] == "rawhide":
                if len(tier[distro]) == 1:
                    continue
                idx = 1
            else:
                idx = 0

            tier1_distro = distro
            tier1_version = tier[distro].pop(idx)
            if not tier[distro]:
                del tier[distro]
            break

if not tier1_distro or not tier1_version:
    print("Warn: no suitable distro for Tier 1 found", file=sys.stderr)

# Always add CentOS Stream at least as Tier 3
for centos_ver_info in distros_info["centos"]:
    version = centos_ver_info["version"]
    found = False

    if tier1_distro == "centos" and tier1_version == version:
        found = True
    for tier in (tier2, tier3):
        if "centos" in tier and version in tier["centos"]:
            found = True
            break

    if not found:
        tier3.setdefault("centos", []).append(version)

# Print the config.yml needed for the corresponding stable branch
branch = "main" if nm_version == "main" else "nm-" + nm_version.replace(".", "-")
print("\n# .gitlab-ci/config.yml for branch '{}'".format(branch))
print(
    """---
# This file contains the configuration for the gitlab ci.
#
# To recreate the .gitlab-ci.yml file, run
#   ci-fairy generate-template
#
# The ci-fairy tool is part of
# https://gitlab.freedesktop.org/freedesktop/ci-templates
#

# Some distros are fairly similar, and we reuse similar scripts.
# The base type maps the distro name to their base.
base_types:
  fedora: fedora
  centos: fedora
  debian: debian
  ubuntu: debian
  alpine: alpine

# The list of all distributions we want to create job for.
distributions:
  # TIER 1: CI run for all MRs.
  # The first tier:1 in the list is used to build the pages and check-{tree,patch}."""
)
print("  - name: {}".format(tier1_distro))
print("    tier: 1")
print("    versions:")
print("      - '{}'".format(tier1_version))

print(
    """
  # TIER 2: distribution versions that will or might use the current NM version.
  # Run when doing a release."""
)
for distro, versions in tier2.items():
    print("  - name: {}".format(distro))
    print("    tier: 2")
    print("    versions:")
    for version in versions:
        print("      - '{}'".format(version))

print(
    """
  # TIER 3: distribution versions not in EOL but don't use the current NM version.
  # Run when doing a release, but a failure won't be blocking for the release."""
)
for distro, versions in tier3.items():
    print("  - name: {}".format(distro))
    print("    tier: 3")
    print("    versions:")
    for version in versions:
        print("      - '{}'".format(version))



#!/usr/bin/env python
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2011 - 2012 Red Hat, Inc.
#

import dbus, socket, struct

# This example lists basic information about network interfaces known to NM

# For the types see include/NetworkManager.h


def int_to_ip(ip_int):
    return socket.inet_ntoa(struct.pack("=I", ip_int))


devtypes = {
    1: "Ethernet",
    2: "Wi-Fi",
    5: "Bluetooth",
    6: "OLPC",
    7: "WiMAX",
    8: "Modem",
    9: "InfiniBand",
    10: "Bond",
    11: "VLAN",
    12: "ADSL",
    13: "Bridge",
    14: "Generic",
    15: "Team",
    16: "TUN",
    17: "IPTunnel",
    18: "MACVLAN",
    19: "VXLAN",
    20: "Veth",
}

states = {
    0: "Unknown",
    10: "Unmanaged",
    20: "Unavailable",
    30: "Disconnected",
    40: "Prepare",
    50: "Config",
    60: "Need Auth",
    70: "IP Config",
    80: "IP Check",
    90: "Secondaries",
    100: "Activated",
    110: "Deactivating",
    120: "Failed",
}

bus = dbus.SystemBus()

# Get a proxy for the base NetworkManager object
proxy = bus.get_object(
    "org.freedesktop.NetworkManager", "/org/freedesktop/NetworkManager"
)
manager = dbus.Interface(proxy, "org.freedesktop.NetworkManager")

# Get all devices known to NM and print their properties
devices = manager.GetDevices()

for d in devices:
    dev_proxy = bus.get_object("org.freedesktop.NetworkManager", d)
    prop_iface = dbus.Interface(dev_proxy, "org.freedesktop.DBus.Properties")
    props = prop_iface.GetAll("org.freedesktop.NetworkManager.Device")

    print("============================")

    print("Interface: %s" % props["Interface"])
    print("Ip 4 Address: %s" % int_to_ip(props["Ip4Address"]))
    print("Ip 4 Config: %s" % props["Ip4Config"])

    try:
        devtype = devtypes[props["DeviceType"]]
    except KeyError:
        devtype = "Unknown"
    print("Type: %s" % devtype)
    print("Driver: %s" % props["Driver"])

    try:
        state = states[props["State"]]
    except KeyError:
        state = "Unknown"
    print("State: %s" % state)
